#!/usr/bin/env node
/*!
 * gitmars v4.0.0-alpha.17
 * 这是一个git工作流工具
 * (c) 2021-2022 saqqdy<https://github.com/saqqdy> 
 * Released under the MIT License.
 */
import { program } from 'commander';
import dayjs from 'dayjs';
import inquirer from 'inquirer';
import chalk from 'chalk';
import getUserToken from '@gitmars/core/lib/api/getUserToken';
import getIsGitProject from '@gitmars/core/lib/git/getIsGitProject';
import getGitConfig from '@gitmars/core/lib/git/getGitConfig';
import sendGroupMessage from '@gitmars/core/lib/sendGroupMessage';
import { createArgs } from '@gitmars/core/lib/utils/command';
import echo from '@gitmars/core/lib/utils/echo';
import getConfig from '@gitmars/core/lib/getConfig';
import { getMergeRequestList, acceptMergeRequest, getMergeRequestChanges, deleteMergeRequest, updateMergeRequest } from '@gitmars/core/lib/api/mergeRequest';
import { getMergeRequestNotesList } from '@gitmars/core/lib/api/mergeRequestNotes';
import i18n from '#lib/locales/index';
import approveConfig from '#lib/conf/approve';

const { blue, cyan, green, magenta, red, yellow } = chalk;
const { args, options } = approveConfig;
if (!getIsGitProject()) {
  echo(red(i18n.__("The current directory is not a git project directory")));
  process.exit(1);
}
const { appName } = getGitConfig();
const config = getConfig();
program.name("gitm approve").usage("[--state [state]] [--quiet]").description(i18n.__("Approve remote merge request"));
if (args.length > 0)
  program.arguments(createArgs(args));
options.forEach((o) => {
  program.option(o.flags, o.description, o.defaultValue);
});
program.action(async (opt) => {
  const userInfoApi = config.apis && config.apis.userInfo && config.apis.userInfo.url || config.api;
  const {
    token,
    level,
    nickname = ""
  } = userInfoApi ? await getUserToken() : {};
  if (level && level > 3) {
    echo(
      red(
        i18n.__("Hey {{name}}, you don't have permission", { nickname })
      )
    );
    process.exit(1);
  }
  const mrList = await getMergeRequestList({ token, state: opt.state });
  if (mrList.length === 0) {
    echo(
      yellow(i18n.__("No merge request record found, process has exited"))
    );
    process.exit(0);
  }
  const prompt = [
    {
      type: "checkbox",
      message: i18n.__("Please select the merge request to be operated"),
      name: "iids",
      choices: []
    },
    {
      type: "list",
      message: i18n.__("Please select the action below."),
      name: "accept",
      choices: [
        i18n.__("View Details"),
        i18n.__("Passed"),
        i18n.__("Not passed"),
        i18n.__("Failed and deleted"),
        i18n.__("Exit")
      ],
      when(answers) {
        return answers.iids.length;
      }
    }
  ];
  for (const mr of mrList) {
    const {
      iid,
      author,
      source_branch,
      target_branch,
      merge_status,
      created_at
    } = mr;
    mr.notes = (await getMergeRequestNotesList({
      token,
      iid
    }) || []).filter((note) => !note.system);
    const disabled = merge_status !== "can_be_merged";
    const _time = dayjs(created_at).format("YYYY/MM/DD HH:mm");
    prompt[0].choices.push({
      name: i18n.__(
        "{{id}} request merge {{{source}}} to {{{target}}} {{disabled}} | {{name}} | {{comments}} | {{time}}",
        {
          id: green(iid + ": "),
          source: green(source_branch),
          target: green(target_branch),
          disabled: disabled ? red(`[ ${i18n.__("Conflict or no need to merge")} ]`) : "",
          name: yellow(author.name),
          comments: green(
            i18n.__("{{length}} comments", {
              length: String(mr.notes.length)
            })
          ),
          time: blue(_time)
        }
      ),
      value: iid,
      checked: false
    });
  }
  const { iids, accept } = await inquirer.prompt(prompt);
  if (iids.length === 0) {
    echo(
      yellow(
        i18n.__("No merge request record selected, process has exited")
      )
    );
    process.exit(0);
  }
  for (const iid of iids) {
    const { merge_status, source_branch, target_branch } = mrList.find(
      (item) => item.iid === iid
    );
    const CAN_BE_MERGED = merge_status === "can_be_merged";
    if (accept === i18n.__("Passed")) {
      if (!CAN_BE_MERGED) {
        echo(
          yellow(
            i18n.__(
              "Requests that can't be merged can't be clicked for review and approval"
            )
          )
        );
        process.exit(0);
      }
      await acceptMergeRequest({ token, iid });
      !opt.quiet && sendGroupMessage(
        i18n.__(
          "{{app}} item {{{source}}} merged to {{{target}}} request ID {{id}} has been merged",
          {
            app: appName,
            source: source_branch,
            target: target_branch,
            id: iid
          }
        )
      );
      echo(green(i18n.__("Merge request {{id}}: Merged", { id: iid })));
    } else if (accept === i18n.__("View Details")) {
      const { changes, changes_count } = await getMergeRequestChanges({
        token,
        iid
      });
      echo(
        green(
          "\n" + i18n.__("{{id}}: total {{total}} files changed", {
            id: iid,
            total: changes_count
          })
        )
      );
      for (const { old_path, new_path, diff } of changes) {
        echo(
          magenta(
            "\n----------------------------------------------------------------------------------"
          )
        );
        echo(magenta(old_path));
        old_path !== new_path && echo(magenta(new_path + `(${i18n.__("New path")})`));
        echo(
          diff.replace(
            /(@@.+)\n/g,
            (m, p1) => cyan(p1) + "\n"
          ).replace(
            /\n(-.+)/g,
            (m, p1) => "\n" + red(p1)
          ).replace(
            /\n(\+.+)/g,
            (m, p1) => "\n" + green(p1)
          )
        );
      }
    } else if (accept === i18n.__("Failed and deleted")) {
      await deleteMergeRequest({ token, iid });
      !opt.quiet && sendGroupMessage(
        i18n.__(
          "{{app}} item {{{source}}} merged to {{{target}}} request ID {{id}} has been deleted",
          {
            app: appName,
            source: source_branch,
            target: target_branch,
            id: iid
          }
        )
      );
      echo(green(i18n.__("Merge request {{id}}: Deleted", { id: iid })));
    } else if (accept === i18n.__("Not passed")) {
      await updateMergeRequest({
        token,
        iid,
        data: { state_event: "close" }
      });
      !opt.quiet && sendGroupMessage(
        i18n.__(
          "{{app}} item {{{source}}} merged to {{{target}}} request ID {{id}} has been closed",
          {
            app: appName,
            source: source_branch,
            target: target_branch,
            id: iid
          }
        )
      );
      echo(green(i18n.__("Merge request {{id}}: Closed", { id: iid })));
    }
  }
});
program.parse(process.argv);
