#!/usr/bin/env node
/*!
 * gitmars v4.0.0-alpha.17
 * 这是一个git工作流工具
 * (c) 2021-2022 saqqdy<https://github.com/saqqdy> 
 * Released under the MIT License.
 */
import { createRequire } from 'node:module';
import { program } from 'commander';
import sh from 'shelljs';
import chalk from 'chalk';
import getType from 'js-cool/es/getType';
import { queue } from '@gitmars/core/lib/queue';
import getIsGitProject from '@gitmars/core/lib/git/getIsGitProject';
import getCurrentBranch from '@gitmars/core/lib/git/getCurrentBranch';
import getGitConfig from '@gitmars/core/lib/git/getGitConfig';
import getIsMergedTargetBranch from '@gitmars/core/lib/git/getIsMergedTargetBranch';
import getIsBranchOrCommitExist from '@gitmars/core/lib/git/getIsBranchOrCommitExist';
import checkGitStatus from '@gitmars/core/lib/git/checkGitStatus';
import searchBranches from '@gitmars/core/lib/git/searchBranches';
import { createArgs } from '@gitmars/core/lib/utils/command';
import { isNeedUpgrade, upgradeGitmars } from '@gitmars/core/lib/versionControl';
import getConfig from '@gitmars/core/lib/getConfig';
import getUserToken from '@gitmars/core/lib/api/getUserToken';
import { defaults } from '#lib/common/global';
import endConfig from '#lib/conf/end';
import i18n from '#lib/locales/index';

const require = createRequire(import.meta.url);
const { args, options } = endConfig;
const { red } = chalk;
if (!getIsGitProject()) {
  sh.echo(
    red(i18n.__("The current directory is not a git project directory"))
  );
  process.exit(1);
}
const config = getConfig();
const { appName } = getGitConfig();
const mergeRequestModule = require.resolve("@gitmars/core/lib/api/mergeRequest");
program.name("gitm end").usage(
  "[type] [name] [--description [description]] [--as-feature] [--no-combine]"
).description(
  i18n.__(
    "Merge bugfix task branch, merge feature function development branch, the corresponding branch will be deleted after the merge is completed"
  )
);
if (args.length > 0)
  program.arguments(createArgs(args));
options.forEach((o) => {
  program.option(o.flags, o.description, o.defaultValue);
});
program.action(
  async (type, name, opt) => {
    const userInfoApi = config.apis && config.apis.userInfo && config.apis.userInfo.url || config.api;
    const needUpgrade = await isNeedUpgrade();
    needUpgrade && upgradeGitmars();
    const allow = ["bugfix", "feature", "support"];
    const deny = [
      defaults.master,
      defaults.develop,
      defaults.release,
      defaults.bugfix,
      defaults.support
    ];
    const {
      token,
      level,
      nickname = ""
    } = userInfoApi ? await getUserToken() : {};
    const status = checkGitStatus();
    let _nameArr = [], isDescriptionCorrect = true;
    if (!status)
      process.exit(1);
    if (config.descriptionValidator) {
      const reg = getType(config.descriptionValidator) === "regexp" ? config.descriptionValidator : new RegExp(config.descriptionValidator);
      isDescriptionCorrect = opt.description && reg.test(opt.description);
    }
    if (!type) {
      [type, ..._nameArr] = getCurrentBranch().split("/");
      name = _nameArr.join("/");
      if (!name) {
        deny.includes(type) && sh.echo(
          red(
            i18n.__(
              "Hey bro, what is the fuck are you doing by executing this command in the {{type}} branch?",
              { type }
            )
          )
        );
        process.exit(1);
      }
    } else if (!name) {
      if (allow.includes(type)) {
        sh.echo(i18n.__("Please enter branch name"));
        process.exit(1);
      }
      const branches = searchBranches({ type });
      if (branches.length === 1) {
        [type, ..._nameArr] = branches[0].split("/");
        name = _nameArr.join("/");
      } else {
        sh.echo(
          branches.length > 1 ? i18n.__(
            "If you find multiple branches with names containing {{type}}, please enter the branch type",
            { type }
          ) : red(
            i18n.__(
              "Branch does not exist, please enter it correctly"
            )
          )
        );
        process.exit(1);
      }
    }
    const isRemoteBranchExist = getIsBranchOrCommitExist(
      `${type}/${name}`,
      true
    );
    if (allow.includes(type) && name) {
      const base = opt.asFeature ? config.release : type === "bugfix" ? config.bugfix : config.release;
      let cmd = [];
      const isNeedCombineDevelop = !getIsMergedTargetBranch(
        `${type}/${name}`,
        config.develop,
        true
      );
      const isNeedCombineBase = !getIsMergedTargetBranch(
        `${type}/${name}`,
        base,
        true
      );
      const isNeedCombineBugfix = !getIsMergedTargetBranch(
        `${type}/${name}`,
        config.bugfix,
        true
      );
      if (opt.combine && isNeedCombineDevelop) {
        cmd = [
          "git fetch",
          `git checkout ${config.develop}`,
          "git pull",
          {
            cmd: `git merge --no-ff ${type}/${name}`,
            config: {
              again: false,
              success: i18n.__(
                "Merge {{{source}}} into {{{target}}} successfully",
                {
                  source: `${type}/${name}`,
                  target: config.develop
                }
              ),
              fail: i18n.__(
                "An error occurred merging {{{source}}} to {{{target}}}, Please follow the instructions",
                {
                  source: `${type}/${name}`,
                  target: config.develop
                }
              )
            }
          },
          {
            cmd: "git push",
            config: {
              again: true,
              success: i18n.__("Successful Pushed"),
              fail: i18n.__(
                "Push failed, please follow the prompts"
              )
            }
          },
          `git checkout ${type}/${name}`
        ];
      }
      if (type === "support" && opt.combine && isNeedCombineBugfix) {
        if (!level || level < 3) {
          cmd = cmd.concat([
            "git fetch",
            `git checkout ${config.bugfix}`,
            "git pull",
            {
              cmd: `git merge --no-ff ${type}/${name}`,
              config: {
                again: false,
                success: i18n.__(
                  "Merge {{{source}}} into {{{target}}} successfully",
                  {
                    source: `${type}/${name}`,
                    target: config.bugfix
                  }
                ),
                fail: i18n.__(
                  "An error occurred merging {{{source}}} to {{{target}}}, Please follow the instructions",
                  {
                    source: `${type}/${name}`,
                    target: config.bugfix
                  }
                )
              }
            },
            {
              cmd: "git push",
              config: {
                again: true,
                success: i18n.__("Successful Pushed"),
                fail: i18n.__(
                  "Push failed, please follow the prompts"
                )
              }
            },
            `git checkout ${type}/${name}`
          ]);
        } else {
          if (!isDescriptionCorrect) {
            sh.echo(
              red(
                i18n.__(
                  "The description of the reason for submission does not meet the specification"
                )
              )
            );
            process.exit(1);
          }
          cmd = cmd.concat([
            {
              cmd: `git push --set-upstream origin ${type}/${name}`,
              config: {
                again: true,
                success: i18n.__(
                  "Push remote and associate remote branch successfully"
                ),
                fail: i18n.__(
                  "Push remote failed, please follow the prompts"
                )
              }
            },
            {
              cmd: {
                module: mergeRequestModule,
                entry: "createMergeRequest",
                options: {
                  source_branch: `${type}/${name}`,
                  target_branch: config.bugfix,
                  token,
                  description: opt.description
                }
              },
              config: {
                again: true,
                success: i18n.__(
                  "Successfully created merge request"
                ),
                fail: i18n.__(
                  "There was an error creating the merge request, please follow the instructions"
                )
              }
            },
            `gitm postmsg "${i18n.__(
              "{{nickname}} submitted a merge request for {{{source}}} branch to {{{target}}} branch in {{app}} project",
              {
                nickname,
                app: appName,
                source: `${type}/${name}`,
                target: config.bugfix
              }
            )}"`
          ]);
        }
      }
      if (!opt.combine || !isNeedCombineBase) {
        cmd = cmd.concat([
          `git checkout ${config.develop}`,
          `git branch -D ${type}/${name}`,
          {
            cmd: "git remote prune origin",
            config: {
              again: true,
              success: i18n.__(
                "Cleanup of remote branch was successful"
              ),
              fail: i18n.__(
                "Failed to clean up remote branch, please follow the prompts"
              )
            }
          }
        ]);
        if (isRemoteBranchExist) {
          cmd = cmd.concat([
            {
              cmd: `git push origin --delete ${type}/${name}`,
              config: {
                again: true,
                success: i18n.__(
                  "Successfully deleted remote branch"
                ),
                fail: i18n.__(
                  "Deletion failed, please contact administrator"
                )
              }
            }
          ]);
        }
      } else {
        if (!level || level < 3) {
          cmd = cmd.concat([
            "git fetch",
            `git checkout ${base}`,
            "git pull",
            {
              cmd: `git merge --no-ff ${type}/${name}`,
              config: {
                again: false,
                success: i18n.__(
                  "Merge {{{source}}} into {{{target}}} successfully",
                  {
                    source: `${type}/${name}`,
                    target: base
                  }
                ),
                fail: i18n.__(
                  "An error occurred merging {{{source}}} to {{{target}}}, Please follow the instructions",
                  {
                    source: `${type}/${name}`,
                    target: base
                  }
                )
              }
            },
            {
              cmd: "git push",
              config: {
                again: true,
                success: i18n.__("Successful Pushed"),
                fail: i18n.__(
                  "Push failed, please follow the prompts"
                )
              }
            },
            `git checkout ${config.develop}`,
            `git branch -D ${type}/${name}`,
            {
              cmd: "git remote prune origin",
              config: {
                again: true,
                success: i18n.__(
                  "Cleanup of remote branch was successful"
                ),
                fail: i18n.__(
                  "Failed to clean up remote branch, please follow the prompts"
                )
              }
            }
          ]);
          if (isRemoteBranchExist) {
            cmd = cmd.concat([
              {
                cmd: `git push origin --delete ${type}/${name}`,
                config: {
                  again: true,
                  success: i18n.__(
                    "Successfully deleted remote branch"
                  ),
                  fail: i18n.__(
                    "Deletion failed, please contact administrator"
                  )
                }
              }
            ]);
          }
        } else {
          if (!isDescriptionCorrect) {
            sh.echo(
              red(
                i18n.__(
                  "The description of the reason for submission does not meet the specification"
                )
              )
            );
            process.exit(1);
          }
          cmd = cmd.concat([
            {
              cmd: `git push --set-upstream origin ${type}/${name}`,
              config: {
                again: true,
                success: i18n.__(
                  "Push remote and associate remote branch successfully"
                ),
                fail: i18n.__(
                  "Push remote failed, please follow the prompts"
                )
              }
            },
            {
              cmd: {
                module: mergeRequestModule,
                entry: "createMergeRequest",
                options: {
                  source_branch: `${type}/${name}`,
                  target_branch: base,
                  token,
                  description: opt.description
                }
              },
              config: {
                again: true,
                success: i18n.__(
                  "Successfully created merge request"
                ),
                fail: i18n.__(
                  "There was an error creating the merge request, please follow the instructions"
                )
              }
            },
            `gitm postmsg "${i18n.__(
              "{{nickname}} submitted a merge request for {{{source}}} branch to {{{target}}} branch in {{app}} project",
              {
                nickname,
                app: appName,
                source: `${type}/${name}`,
                target: base
              }
            )}"`
          ]);
        }
      }
      queue(cmd);
    } else {
      sh.echo(
        red(
          i18n.__("type only allows input") + ": " + JSON.stringify(allow)
        )
      );
      process.exit(1);
    }
  }
);
program.parse(process.argv);
